#!/usr/bin/env python
# -*- coding: utf-8 -*-
########################################################################
#
# mamebrowse.py
# A webserver for reading and browsing MAME's history.dat
#
# MAMEbrowse (c) 2011 Elias Schwerdtfeger, http://www.tamagothi.de/
#
# Licensed under the terms of the pirates' license
# Share and enjoy, but don't sue me!
#
########################################################################

import BaseHTTPServer
import codecs
import os
import os.path
import sys
import stat
import datetime
import random
import re
from cgi import escape
from optparse import OptionParser


########################################################################
#
# Configuration
# This is the part you want to change.
# 
########################################################################

HISTORY_NAME = 'history.dat'
HISTORY_DIR = '/usr/share/games/xmame/'
HISTORY_ENCODING = 'iso-8859-1'
SERVER_PORT = 8000


########################################################################
#
# Global variables
#
########################################################################

PROGNAME = 'MAMEBrowse'
VERSION_NUMBER = '0.1'
VERSION_ID = '$Id: mamebrowse.py,v 1.6 2011/01/04 06:52:59 elias Exp $'
VERSION_STRING = PROGNAME + '/' + VERSION_NUMBER
contents = None


########################################################################
#
# CSS Stylesheet, statically coded
# (The whole thing is in one file, simple, but harder to change)
#
########################################################################

CSS_STYLE= """\
body {
background-color:#000;
color:#dde;
font-family:Arial,Hevetica,sans-serif;
font-size:10pt;
margin:0;
padding:.5em;
}
a:link,a:visited {
color:#8ae;
text-decoration:none;
}
a:active,a:hover {
color:#ea8;
text-decoration:underline;
}
.wrap {
background-color:#222;
width:780px;
margin:0 auto;
border:2px solid #112;
}
.head,.foot {
color:#ccd;
padding:0;
margin:0;
}
.head {
background-color:#445;
text-align:center;
}
.foot {
background-color:#112;
}
.head h1 {
margin:0;
padding-top:40px;
padding-bottom:10px;
font-size:140%;
font-weight:normal;
}
.foot {
text-align:right;
padding-right:4px;
font-size:80%;
clear:both;
}
.side,.main {
margin:0;
padding-top:.8em;
margin-bottom:.8em;
}
.side {
float:left;
top:0px;
width:190px;
padding-left:.5em;
padding-right: 8px;
border-right:1px dotted #888;
}
.main {
margin-left:220px;
top:0px;
width:540px;
font-size:120%;
}
.main p, .main h2, .main h3 {
margin: 0;
line-height: 120%;
margin-bottom: 0.6em;
}
.main h2 {
font-size:120%;
font-weight:normal;
}
.main h2 small {
font-size:60%;
color:#99a;
}
.main h3 {
font-size: 120%;
font-weight:normal;
margin-top: 0.4em;
background-color:#333;
}
.tnav {
background-color:#334;
text-align:left;
}
.snav {
background-color:#223;
text-align:left;
}
.nav ul, .side ul {
list-style:none;
margin:0;
padding:0;
}
.nav ul li {
display:inline;
padding:0;
margin:0;
margin-left:.6em;
}
.side ul li {
margin:0;
padding:0;
}
p.r {
font-family:'Courier New',Courier,monospace;
border-bottom:1px dotted #888;
padding-bottom:0.3em;
}
p.r:before {
content:'ROMs: ';
}
form {
padding-bottom:1em;
border-bottom:1px dotted #888;
}
input {
background-color:#000;
color:#eef;
padding:2px;
border:1px solid #888;
}
"""


########################################################################
#
# Class of a history.dat entry
#
########################################################################

class Entry:
    re_fine = re.compile(r'[^0-9a-z]+')
    re_nosp_1 = re.compile(r'^-*')
    re_nosp_2 = re.compile(r'-*$')
    re_copy = re.compile(r'(\(c\).*)$', re.IGNORECASE | re.UNICODE)

    def __init__(self, c_instance, name, roms, f_pos, f_to):
        if f_to <= f_pos:
            raise RuntimeError, '%s pos=%d, to=%d' % (name, f_pos, f_to)
        self.__cinst = c_instance
        self.game_name = name
        tmp = self.re_fine.sub('-', name.lower())
        tmp = self.re_nosp_1.sub('', tmp);
        self.short_name = self.re_copy.sub('', self.game_name)
        self.nice_name = self.re_nosp_2.sub('', tmp);
        self.klass = self.nice_name[0]
        if self.klass.isdigit():
            self.klass = '0-9'
        self.rom_names = roms
        self.__f_start_pos = f_pos
        self.__f_end_pos = f_to

    def get_raw_text(self):
        f = open(self.__cinst.filepath, 'rt')
        try:
            f.seek(self.__f_start_pos)
            tmp = f.read(self.__f_end_pos - self.__f_start_pos)
        finally:
            f.close()
        return tmp.decode(HISTORY_ENCODING, 'ignore')

    re_emptyln = re.compile(r'(\n\s*)+', re.UNICODE)
    re_title = re.compile(r'\-\s+(.*)\s+\-\s*$', re.UNICODE)
    re_ul = re.compile(r'\s*\*(.*)$', re.UNICODE)

    def get_html_text(self):
        t = self.get_raw_text()
        t = self.re_emptyln.sub('\n', t)
        h = []
        first_line = True
        ul_active = False
        for l in t.split('\n'):
            l = l.strip()
            if len(l) == 0:
                continue
            l = escape(l)
            if first_line:
                l = self.re_copy.sub(ur'<small>\1</small>', l)
                h.append(u'<h2>%s</h2>' % (l,))
                h.append(u'<p class="r">%s</p>' % (', '.join(self.rom_names)))
                first_line = False
                continue
            mo = self.re_title.match(l)
            if mo:
                if ul_active:
                    h.append(u'</ul>')
                    ul_active = False
                h.append(u'<h3>%s</h3>' % (mo.group(1),))
                continue
            mo = self.re_ul.match(l)
            if mo:
                if not ul_active:
                    h.append(u'<ul>')
                    ul_active = True
                h.append(u'<li>%s</li>' % (mo.group(1),))
                continue
            if ul_active:
                h.append(u'</ul>')
                ul_active = False
            h.append(u'<p>%s</p>' % (l,))
        if ul_active:
            h.append(u'</ul>')
        return u'\n'.join(h)



########################################################################
#
# The class for the singleton contents object
# TODO: Optimize 
#
########################################################################

class ContentsClass:
    def __init__(self, filepath, callback=None, ecls=Entry):
        self.filepath = filepath
        st = os.stat(self.filepath)
        stctime = st[stat.ST_CTIME]
        self.filetime_t = datetime.datetime.fromtimestamp(stctime)
        self.filetime = self.filetime_t.isoformat(' ')
        self.entries = []
        self.rom_entries = {}
        self.name_entries = {}
        self.nice_entries = {}
        self.klass_entries = {}
        self.klass_array = []
        self.number_of_entries = 0
        self.number_of_roms = 0
        self.number_of_warnings = 0
        self.warnings = []
        self.__create_entries(callback, ecls)

    def __warn(self, text, name):
        tmp = '%s "%s"' % (text, name)
        self.warnings.append(tmp)

    def __create_entries(self, callback, ecls):
        rom_names = None
        game_name = None
        count = 0
        pos = 0
        lastpos = 0
        state = 0
        b_begin = 0
        b_end = 0
        fo = open(self.filepath, 'rt')
        try:
            while True:
                ln = fo.readline()
                if not ln:
                    break
                lastpos = pos
                pos = fo.tell()
                ln = ln.decode(HISTORY_ENCODING).strip()
                if state == 0:
                    if ln.startswith('$info='):
                        tmp = ln[6:]
                        rom_names = [i for i in tmp.split(',') if len(i) > 0]
                        state = 1
                elif state == 1:
                    if ln.startswith('$bio'):
                        state = 2
                elif state == 2:
                    if len(ln) > 0:
                        b_begin = lastpos
                        game_name = ln
                        state = 3
                elif state == 3:
                    if ln.startswith('$end'):
                        b_end = lastpos
                        e = ecls(self, game_name, rom_names, b_begin, b_end)
                        self.entries.append(e)
                        if (callback):
                            count += 1
                            callback(count)
                        if self.name_entries.has_key(game_name):
                            self.__warn('Double game', game_name)
                            state = 0
                        else:
                            self.name_entries[game_name] = e
                        for tmp in rom_names:
                            if self.rom_entries.has_key(tmp):
                                self.__warn('Double rom', tmp)
                            else:
                                self.rom_entries[tmp] = e
                        if self.klass_entries.has_key(e.klass):
                            self.klass_entries[e.klass].append(e)
                        else:
                            self.klass_entries[e.klass] = [e]
                        if self.nice_entries.has_key(e.nice_name):
                            self.__warn('Double nice name', e.nice_name)
                        else:
                            self.nice_entries[e.nice_name] = e
                        state = 0
        finally:
            fo.close()
        self.number_of_entries = len(self.name_entries)
        self.number_of_roms = len(self.rom_entries)
        self.number_of_warnings = len(self.warnings)
        self.klass_array = self.klass_entries.keys()
        self.klass_array.sort()


########################################################################
#
# Create the global contents instance
#
########################################################################

def create_contents(path, cls=ContentsClass, ecls=Entry):
    global contents
    if contents == None:
        if os.path.isdir(path):
            lpath = os.path.join(path, HISTORY_NAME)
            if not os.path.exists(path):
                raise RuntimeError, 'No MAME history file %s' % (lpath,)
        contents = cls(path)


########################################################################
#
# The server class for mamebrowse
# (Isn't anything special for now, just a plain HTTP server)
#
########################################################################

class Server(BaseHTTPServer.HTTPServer):
    pass


########################################################################
#
# The server response class for mamebrowse
# (This is the real worker in the server)
#
########################################################################

class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
    server_version = VERSION_STRING
    protocol_version = 'HTTP/1.0'

    def wline(self, line, *args):
        l = line % args
        self.cfile.write(l + '\n')

    def link(self, text, href, title=None):
        if title != None:
            return u'<a href="%s" title="%s">%s</a>' % (href, title, text)
        else:
            return u'<a href="%s">%s</a>' % (href, text)

    def wlink(self, text, href, title=None):
        self.wline(self.link(text, href, title))

    def wllink(self, text, href, title=None):
        self.wline('<li>%s</li>', self.link(text, href, title))

    def wglink(self, e):
        self.wllink(escape(e.short_name), '/game/' + e.nice_name + '/')
        
    def whead(self, title, code=200):
        l = self.wline
        self.send_response(code)
        self.send_header('Content-Type', 'application/xhtml+xml;charset=utf-8')
        self.end_headers()
        l(u'<?xml version="1.0" ?>')
        l(u'<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"')
        l(u' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">')
        l(u'<html xmlns="http://www.w3.org/1999/xhtml">')
        l(u'<head>')
        l(u'<meta name="generator" content="%s" />', self.server_version)
        l(u'<link rel="stylesheet" type="text/css" href="/style/" />')
        l(u'<title>%s: %s</title>', PROGNAME, escape(title))
        l(u'</head>')
        l(u'<body>')
        l(u'<div class="wrap">')
        l(u'<div class="head">')
        l(u'<h1>%s</h1>', escape(title))
        l(u'</div>')
    
    def wtnav(self):
        ln = self.wline
        lk = self.wllink
        ln(u'<div class="tnav nav">')
        ln(u'<ul>')
        lk(u'Home', u'/', u'Home page')
        lk(u'Random game', u'/random/', u'Random game')
        lk(u'Search game', u'/search/', u'Search game')
        ln(u'</ul>')
        ln(u'</div>')

    def wsnav(self, aktklass='', empty=False):
        ln = self.wline
        lk = self.wllink
        ln(u'<div class="snav nav">')
        ln(u'<ul>')
        if not empty:
            for k in contents.klass_array:
                if k == aktklass:
                    ln(u'<li>%s</li>', k.upper())
                else:
                    l = u'/letter/%s/' % (k,)
                    lk(k.upper(), l)
        ln(u'</ul>')
        ln(u'</div>')

    def wsbar(self, aktklass='', aktentry=None):
        ln = self.wline
        lk = self.wglink
        ln(u'<div class="side">')
        ln(u'<ul>')
        if aktklass != '':
            ea = contents.klass_entries[aktklass]
            for e in ea:
                if e != aktentry:
                    lk(e)
                else:
                    ln(u'<li>%s</li>', escape(e.short_name))
        ln(u'</ul>')
        ln(u'</div>')

    def wfoot(self):
        l = self.wline
        l(u'<div class="foot">')
        l(u'%s', self.server_version)
        l(u'</div>')
        l(u'</div>')
        l(u'</body>')
        l(u'</html>')

    re_sterm = re.compile(r'\?s\=(.*)$')
    re_prorp = re.compile(r'\%\d*')

    def do_GET(self):
        ccls = codecs.getwriter('ascii')
        self.cfile = ccls(self.wfile, 'xmlcharrefreplace')
        cmd = ''
        a = self.path.split('/')
        if len(a) > 1:
            cmd = a[1]
        if cmd in ('home', ''):
            self.dir_home()
        elif cmd in ('style', ):
            self.dir_style()
        elif cmd in ('random', ):
            self.dir_random()
        elif cmd in ('search', ):
            smo = self.re_sterm.search(a[-1])
            if smo:
                term = smo.group(1)
                term.replace('+', ' ')
                term = self.re_prorp.sub('', term)
            else:
                term = ''
            self.dir_search(term)
        elif cmd in ('game', ):
            if len(a) > 2:
                nn = a[2].strip()
                if nn != '':
                    self.dir_game(nn)
                else:
                    self.send_response(303, 'Empty game requested')
                    self.send_header('Location', '/')
                    self.end_headers()
            else:
                self.send_response(303, 'Empty game requested')
                self.send_header('Location', '/')
                self.end_headers()
        elif cmd in ('letter', ):
            if len(a) > 2:
                lt = a[2].strip()
                self.dir_letter(lt)
            else:
                self.dir_home()
        else:
            self.send_error(404)

    def dir_random(self):
        game_name = random.choice(contents.nice_entries.keys())
        ln = '/game/' + game_name + '/'
        self.send_response(303, 'Redirecting to random game')
        self.send_header('Location', ln)
        self.end_headers()

    def dir_search(self, term):
        f = []
        if term != '':
            t = term.lower()
            for e in contents.entries:
                if e.game_name.lower().find(t) >= 0:
                    f.append(e)
            if len(f) == 1:
                ln = '/game/' + f[0].nice_name + '/'
                self.send_response(303, 'Exactly one matching game')
                self.send_header('Location', ln)
                self.end_headers()
                return
        l = self.wline
        self.whead('Search games')
        self.wtnav()
        self.wsnav('')
        l(u'<div class="main">')
        l(u'<h2>Search game</h2>')
        l(u'<form action="/search/" method="get">')
        l(u'<input name="s" value="%s" length="20" />', escape(term))
        l(u'<input type="submit" value="Search game" />')
        l(u'</form>')
        if term != '':
            l(u'<h3>Search for %s</h3>', escape(term))
            if len(f) == 0:
                l(u'<p>No matching game found for %s</p>', escape(term))
            else:
                l(u'<ul>')
                for e in f:
                    self.wglink(e)
                l(u'</ul>')
        else:
            l(u'<p>Search in the games\' names</p>')
        l(u'</div>')
        self.wfoot()

    def dir_style(self):
        self.send_response(200)
        self.send_header('Content-Type', 'text/css')
        self.end_headers()
        self.wfile.write(CSS_STYLE)

    def dir_home(self):
        l = self.wline
        self.whead('Home')
        self.wtnav()
        self.wsnav('')
        l(u'<div class="main">')
        l(u'<h2>Mamehist file</h2>')
        l(u'<p>')
        l(u'Path: <code>%s</code><br />', contents.filepath)
        l(u'Number of entries: %d<br />', contents.number_of_entries)
        l(u'Number of ROMs: %d<br />', contents.number_of_roms)
        l(u'File last changed: %s', contents.filetime)
        l(u'</p>')
        if contents.number_of_warnings > 0:
            l(u'<h3>%d Warnings </h3>', contents.number_of_warnings)
            l(u'<ol>')
            for i in contents.warnings:
                l(u'<li>%s</li>', escape(i))
            l(u'</ol>')
        l(u'</div>')
        self.wfoot()

    def dir_game(self, nice):
        try:
            e = contents.nice_entries[nice]
        except:
            self.send_error(404)
            return
        l = self.wline
        self.whead(e.short_name)
        self.wtnav()
        self.wsnav(e.klass)
        self.wsbar(e.klass, e)
        l('<div class="main">')
        self.cfile.write(e.get_html_text())
        l('</div>')
        self.wfoot()
    
    def dir_letter(self, lt):
        l = self.wline
        self.whead('Letter %s' % (escape(lt),))
        self.wtnav()
        self.wsnav(lt)
        self.wsbar(lt, None)
        l('<div class="main">')
        l('<p>Select game or another letter.</p>')
        l('</div>')
        self.wfoot()


########################################################################
#
# Prepare data and start the server
#
########################################################################

def main():
    ps = OptionParser('Usage: %prog [options]', 
                      version='%prog ' + VERSION_NUMBER)
    ps.add_option('-f', '--file', dest='mamehist', 
                  help='Path to the mamehist.dat file',
                  action='store', type='string',
                  default=os.path.join(HISTORY_DIR, HISTORY_NAME),
                  metavar='File')
    ps.add_option('-p', '--port', dest='port',
                  help='Port for the server to listen',
                  action='store', type='int',
                  default=SERVER_PORT,
                  metavar='Port')
    (opts, args) = ps.parse_args()
    if len(args) > 0:
        sys.stderr.write('Warning: Arguments ignored\n')
    sys.stderr.write('Reading and indexing mamehist.dat...\n')
    create_contents(opts.mamehist)
    sys.stderr.write('Starting the server...\n')
    server_addr = ('', opts.port)
    server = Server(server_addr, Handler)
    sys.stderr.write('Server is up and listening on port %d\n' % (opts.port,))
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        sys.stderr.write('Server terminated\n')


########################################################################
#
# Make this file executable
#
########################################################################

if __name__ == '__main__':
    main()


########################################################################
# 
# May the meek and the downtrodden someday inherit the earth, but for
# now I have to kill some colorful insectoid galaga aliens...
#
# Elias ;)
#
########################################################################
